import logging
import re
from typing import Optional

from semver import VersionInfo

from daktari.check import Check, CheckResult
from daktari.command_utils import get_stdout
from daktari.os import OS
from daktari.version_utils import try_parse_semver


class TfenvInstalled(Check):
    name = "tfenv.installed"

    suggestions = {
        OS.OS_X: "<cmd>brew install tfenv</cmd>",
        OS.UBUNTU: """
            <cmd>git clone https://github.com/tfutils/tfenv.git ~/.tfenv</cmd>
            <cmd>sudo ln -s ~/.tfenv/bin/* /usr/local/bin</cmd>
            """,
        OS.GENERIC: "Install tfenv: https://github.com/tfutils/tfenv",
    }

    def check(self) -> CheckResult:
        return self.verify_install("tfenv")


class TerraformInstalled(Check):
    def __init__(
        self, required_version: Optional[str] = None, recommended_version: Optional[str] = None, use_tfenv: bool = False
    ):
        self.required_version = required_version
        self.recommended_version = recommended_version
        self.name = "terraform.installed"
        self.use_tfenv = use_tfenv

        if use_tfenv:
            self.depends_on = [TfenvInstalled]

            # Read the required tf-version from tfenv (i.e. .terraform-version files)
            tfenv_version = open(".terraform-version", "r").read().strip()
            logging.debug(f"Terraform version required from .terraform-version file is: {tfenv_version}")
            self.required_version = f"=={tfenv_version}"
            self.suggestions = {OS.GENERIC: "<cmd>tfenv install</cmd>"}
        else:
            self.suggestions = {
                OS.OS_X: "<cmd>brew tap hashicorp/tap && brew install hashicorp/tap/terraform</cmd>",
                OS.GENERIC: "Install Terraform: https://learn.hashicorp.com/tutorials/terraform/install-cli",
            }

    def check(self) -> CheckResult:
        installed_version = get_terraform_version()
        return self.validate_semver_expression(
            "Terraform", installed_version, self.required_version, self.recommended_version
        )


version_pattern = re.compile("Terraform v([0-9\\.]+)")


def get_terraform_version() -> Optional[VersionInfo]:
    raw_version = get_stdout("terraform version")
    if raw_version:
        match = version_pattern.search(raw_version)
        if match:
            version = try_parse_semver(match.group(1))
            logging.debug(f"Terraform version: {version}")
            return version
    return None



import logging
from typing import List, Optional

from semver import VersionInfo

from daktari.check import Check, CheckResult
from daktari.command_utils import get_stdout, run_command
from daktari.os import OS
from daktari.version_utils import try_parse_semver


def get_nodejs_version() -> Optional[VersionInfo]:
    version_output = get_stdout("node --version")
    version_output = None if version_output is None else version_output.lstrip("v")
    return try_parse_semver(version_output)


def run_nvm(nvm_args: List[str]):
    return run_command(["sh", "-u", "-c", '. "$NVM_DIR/nvm.sh"; nvm "$@"', "--", *nvm_args])


def can_run_nvm() -> bool:
    try:
        run_nvm(["--version"])
        return True
    except Exception:
        logging.debug("Exception running nvm", exc_info=True)
        return False


def get_nvm_stdout(nvm_args: List[str]) -> Optional[str]:
    try:
        return run_nvm(nvm_args).stdout
    except Exception:
        return None


def nvm_resolve_version(version: str) -> Optional[VersionInfo]:
    resolved_version = get_nvm_stdout(["version", version])
    resolved_version = None if resolved_version is None else resolved_version.lstrip("v")
    return try_parse_semver(resolved_version)


def get_nvmrc_version() -> Optional[VersionInfo]:
    try:
        with open(".nvmrc", "r") as nvmrc_file:
            nvmrc_ver = nvmrc_file.readline().strip()
        return nvmrc_ver
    except IOError:
        logging.debug("Could not read .nvmrc", exc_info=True)
        return None


class NodeJsVersion(Check):
    name = "nodejs.version"

    def __init__(self, required_version: str, recommended_version: Optional[str] = None):
        self.required_version = required_version
        self.recommended_version = recommended_version

    suggestions = {
        OS.GENERIC: "Install node.js",
    }

    def check(self) -> CheckResult:
        nodejs_version = get_nodejs_version()
        logging.info(f"node.js version: {nodejs_version}")
        return self.validate_semver_expression(
            "node.js", nodejs_version, self.required_version, self.recommended_version
        )


class NvmInstalled(Check):
    name = "nvm.installed"

    suggestions = {OS.GENERIC: "Install nvm from: https://nvm.sh"}

    def check(self) -> CheckResult:
        return self.verify(can_run_nvm(), "nvm is <not/> installed")


class NodeJsVersionMatchesNvmrc(Check):
    name = "nodejs.nvmrc.version"
    depends_on = [NvmInstalled]

    suggestions = {
        OS.GENERIC: """
            Run: <cmd>nvm install</cmd> (automatically picks up the right version from .nvmrc)
            Run: <cmd>nvm use</cmd>
            Run: <cmd>nvm alias default <version></cmd> to set the default node version to the newly
            installed one
            """
    }

    def check(self) -> CheckResult:
        nvmrc_version = get_nvmrc_version()
        if nvmrc_version is None:
            return self.failed("Missing or invalid .nvmrc file")

        expected_version = nvm_resolve_version(nvmrc_version)
        if expected_version is None:
            return self.failed(f'node.js version "{nvmrc_version}" is not installed')

        active_version = get_nodejs_version()
        if active_version != expected_version:
            return self.failed(f'the active node.js version is {active_version}, "{nvmrc_version}" is required')

        return self.passed(f"node.js version is {active_version}")



import logging
from typing import Dict, Optional

from python_hosts import Hosts
from tabulate import tabulate

from daktari.check import Check, CheckResult
from daktari.os import OS, check_env_var_exists, get_env_var_value
from daktari.version_utils import get_simple_cli_version


class WatchmanInstalled(Check):
    name = "watchman.installed"

    suggestions = {
        OS.OS_X: "<cmd>brew install watchman</cmd>",
        OS.GENERIC: "Install watchman: https://facebook.github.io/watchman/docs/install.html#buildinstall",
    }

    def check(self) -> CheckResult:
        return self.verify_install("watchman")


class MkcertInstalled(Check):
    name = "mkcert.installed"

    suggestions = {
        OS.OS_X: """
            Install mkcert:
            <cmd>brew install mkcert</cmd>
            Install the local CA in the system trust store:
            <cmd>mkcert -install</cmd>
            """,
        OS.GENERIC: """
            Install mkcert: https://mkcert.dev/#installation
            Install the local CA in the system trust store:
            <cmd>mkcert -install</cmd>
            """,
    }

    def check(self) -> CheckResult:
        return self.verify_install("mkcert")


class KtlintInstalled(Check):
    name = "ktlint.installed"

    suggestions = {
        OS.OS_X: "<cmd>brew install ktlint</cmd>",
        OS.GENERIC: "Install ktlint: https://github.com/pinterest/ktlint#installation",
    }

    def __init__(self, required_version: Optional[str] = None, recommended_version: Optional[str] = None):
        self.required_version = required_version
        self.recommended_version = recommended_version

    def check(self) -> CheckResult:
        installed_version = get_simple_cli_version("ktlint")
        return self.validate_semver_expression(
            "ktlint", installed_version, self.required_version, self.recommended_version
        )


class JqInstalled(Check):
    name = "jq.installed"

    suggestions = {
        OS.OS_X: "<cmd>brew install jq</cmd>",
        OS.UBUNTU: "<cmd>sudo apt install jq</cmd>",
        OS.GENERIC: "Install jq: https://stedolan.github.io/jq/download/",
    }

    def check(self) -> CheckResult:
        return self.verify_install("jq")


class FlywayInstalled(Check):
    name = "flyway.installed"

    suggestions = {
        OS.OS_X: "<cmd>brew install flyway</cmd>",
        OS.UBUNTU: "<cmd>snap install flyway</cmd>",
        OS.GENERIC: "Install flyway: https://flywaydb.org/documentation/usage/commandline/#download-and-installation",
    }

    def check(self) -> CheckResult:
        return self.verify_install("flyway", "-v")


class ShellcheckInstalled(Check):
    name = "shellcheck.installed"

    suggestions = {
        OS.OS_X: "<cmd>brew install shellcheck</cmd>",
        OS.UBUNTU: "<cmd>sudo apt install shellcheck</cmd>",
        OS.GENERIC: "Install shellcheck: https://github.com/koalaman/shellcheck#user-content-installing",
    }

    def check(self) -> CheckResult:
        return self.verify_install("shellcheck")


class MakeInstalled(Check):
    name = "make.installed"

    suggestions = {
        OS.OS_X: "<cmd>xcode-select --install</cmd>",
        OS.UBUNTU: "<cmd>sudo apt install make</cmd>",
        OS.GENERIC: "Install make: https://www.gnu.org/software/make/",
    }

    def check(self) -> CheckResult:
        return self.verify_install("make")


class GccInstalled(Check):
    name = "gcc.installed"

    suggestions = {
        OS.OS_X: "<cmd>xcode-select --install</cmd>",
        OS.UBUNTU: "<cmd>sudo apt install gcc</cmd>",
        OS.GENERIC: "Install gcc: https://gcc.gnu.org/",
    }

    def check(self) -> CheckResult:
        return self.verify_install("gcc")


class EnvVarSet(Check):
    def __init__(self, variable_name: str, variable_value: Optional[str] = "", provision_command: str = ""):
        self.name = f"env.variableSet.{variable_name}"
        self.suggestions = {OS.GENERIC: provision_command}
        self.variable_name = variable_name
        self.variable_value = variable_value

    def check(self) -> CheckResult:
        if self.variable_value:
            return self.verify(
                get_env_var_value(self.variable_name) == self.variable_value,
                f"{self.variable_name} has <not/> got the required value of f{self.variable_value}",
            )
        else:
            return self.verify(check_env_var_exists(self.variable_name), f"{self.variable_name} is <not/> set")


class ShfmtInstalled(Check):
    name = "shfmt.installed"

    suggestions = {
        OS.OS_X: "<cmd>brew install shfmt</cmd>",
        OS.UBUNTU: "<cmd>snap install shfmt</cmd>",
        OS.GENERIC: "Install shfmt: https://github.com/mvdan/sh",
    }

    def check(self) -> CheckResult:
        return self.verify_install("shfmt")


class HostAliasesConfigured(Check):
    name = "hostAliases.configured"

    def __init__(self, required_aliases: Dict[str, str]):
        self.required_aliases = required_aliases
        hosts_path = Hosts.determine_hosts_path()
        entries_text = tabulate([(addr, name) for (name, addr) in self.required_aliases.items()], tablefmt="plain")
        self.suggestions = {OS.GENERIC: f"Add the following entries to {hosts_path}:\n\n{entries_text}"}

    def check(self) -> CheckResult:
        hosts = Hosts()
        entries = [e for e in hosts.entries if e.entry_type in ("ipv4", "ipv6")]
        entries_dict = {}
        for entry in entries:
            for name in entry.names:
                entries_dict[name] = entry.address
        logging.debug(f"Hosts file entries: {entries_dict}")
        for (name, address) in self.required_aliases.items():
            if entries_dict.get(name) != address:
                return self.failed(f"{hosts.hosts_path} alias {name} -> {address} not present")

        return self.passed(f"{hosts.hosts_path} aliases present")


class DetektInstalled(Check):
    name = "detekt.installed"

    def __init__(
        self,
        required_version: Optional[str] = None,
        recommended_version: Optional[str] = None,
        install_version: Optional[str] = None,
    ):
        self.install_version = install_version
        self.required_version = required_version
        self.recommended_version = recommended_version
        self.suggestions = {
            OS.OS_X: self.get_install_cmd(),
            OS.UBUNTU: self.get_install_cmd(),
            OS.GENERIC: "Install detekt: https://detekt.dev/cli.html#install-the-cli",
        }

    def get_install_cmd(self) -> str:
        version = self.install_version or "[desired version - see https://github.com/detekt/detekt/releases]"
        return f"""{{
  DETEKT_VERSION={version}; \\
  mkdir -p ~/.local/bin &&\\
  cd ~/.local/bin &&\\
  curl -sSLO https://github.com/detekt/detekt/releases/download/v$DETEKT_VERSION/detekt-cli-$DETEKT_VERSION.zip &&\\
  unzip detekt-cli-$DETEKT_VERSION.zip &&\\
  rm detekt-cli-$DETEKT_VERSION.zip &&\\
  chmod +x detekt-cli-$DETEKT_VERSION/bin/detekt-cli &&\\
  ln -s ~/.local/bin/detekt-cli-$DETEKT_VERSION/bin/detekt-cli ~/.local/bin/detekt
}}"""

    def check(self) -> CheckResult:
        installed_version = get_simple_cli_version("detekt")
        return self.validate_semver_expression(
            "detekt", installed_version, self.required_version, self.recommended_version
        )



import logging
import re
from typing import Optional

from semver import VersionInfo

from daktari.check import Check, CheckResult
from daktari.command_utils import get_stderr, run_command
from daktari.os import OS

java_version_pattern = re.compile('^.*version "(.*?)".*$', re.MULTILINE)
javac_version_pattern = re.compile("^javac (.*)$", re.MULTILINE)

one_dot_pattern = re.compile("1\\.([0-9]+)")
other_pattern = re.compile("([0-9]+)")


def get_java_version() -> Optional[VersionInfo]:
    version_output = get_stderr("java -version")
    return parse_java_version_output(version_output)


def get_jdk_version() -> Optional[VersionInfo]:
    try:
        version_output = run_command("javac -version")
    except Exception:
        return None

    return parse_javac_version_output(version_output.stdout + version_output.stderr)


def parse_java_version_output(version_output: Optional[str]) -> Optional[VersionInfo]:
    if version_output:
        match = java_version_pattern.search(version_output)
        if match:
            version_string = match.group(1)
            logging.debug(f"Java version string: {version_string}")
            return parse_java_version_string(version_string)
    return None


def parse_javac_version_output(version_output: Optional[str]) -> Optional[VersionInfo]:
    if version_output:
        match = javac_version_pattern.search(version_output)
        if match:
            version_string = match.group(1)
            logging.debug(f"JDK version string: {version_string}")
            return parse_java_version_string(version_string)
    return None


def parse_java_version_string(version_string: str) -> Optional[VersionInfo]:
    try:
        return VersionInfo.parse(version_string)
    except ValueError:
        return parse_alternative_java_version_numbers(version_string)


def parse_alternative_java_version_numbers(version_string: str) -> Optional[int]:
    one_dot_match = one_dot_pattern.search(version_string)
    if one_dot_match:
        return VersionInfo(int(one_dot_match.group(1)))
    other_pattern_match = other_pattern.search(version_string)
    if other_pattern_match:
        return VersionInfo(int(other_pattern_match.group(1)))
    return None


class JavaVersion(Check):
    name = "java.version"

    def __init__(
        self,
        required_version: Optional[str] = None,
        recommended_version: Optional[str] = None,
        java_instructions: Optional[str] = None,
    ):
        self.required_version = required_version
        self.recommended_version = recommended_version

        java_instructions = f"\n\n{java_instructions}" if java_instructions else ""
        self.suggestions = {OS.GENERIC: f"""Install Java{java_instructions}"""}

    def check(self) -> CheckResult:
        java_version = get_java_version()
        logging.info(f"Java version: {java_version}")
        return self.validate_semver_expression("Java", java_version, self.required_version, self.recommended_version)


class JdkVersion(Check):
    name = "jdk.version"

    def __init__(self, required_version: str, recommended_version: Optional[str] = None):
        self.required_version = required_version
        self.recommended_version = recommended_version

    suggestions = {
        OS.GENERIC: "Install Java JDK",
    }

    def check(self) -> CheckResult:
        jdk_version = get_jdk_version()
        logging.info(f"JDK version: {jdk_version}")
        return self.validate_semver_expression("JDK", jdk_version, self.required_version, self.recommended_version)
